/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.example.warehouse.manager.service;

import java.text.DateFormat;
import java.util.Date;

import org.bluestemsoftware.open.eoa.example.warehouse.manager.service.dao.CustomerOrderDAO;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.service.dao.DAOException;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.Constants;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.OutstandingRequest;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.WarehouseError;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.CustomerOrder;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.CustomerOrder.Status;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplication;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplicationContext;
import org.bluestemsoftware.specification.eoa.application.spring.MyOperation;
import org.bluestemsoftware.specification.eoa.application.spring.MyRole;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerApplication;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerOperation;
import org.bluestemsoftware.specification.eoa.application.spring.PartnerRole;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.w3c.dom.Element;

/**
 * Encapsulates logic related to order fulfillment within context of warehouse manager
 * application. Note that this service exposes a web-service interface.
 */
public class OrderProcessor implements Service {

    public static String ROLE_NAME = "http://mycompany.com/wm/order/processor";

    private ShippingProcessor shippingProcessor;
    private CustomerOrderDAO customerOrderDAO;
    private MyRole myRole;

    /**
     * This method maps to 'scheduleOrder' operation defined on interface implemented by this
     * web service. Invoked by <code>SpringEngine</code>. Schedules an existing order
     * 'promised' to customer by {@link AvailabilityProcessor} by updating order status to
     * 'scheduled' and scheduling the order for shipment. Note that this method demonstrates
     * the handling of a potentially long-running message exchange.
     * 
     * @param myOperation
     * @param requestContext
     */
    public void scheduleOrder(MyOperation myOperation, ActionContext requestContext) {

        // using a message property (which maps a schema type to a
        // predefined query string) we extract the zipcode from
        // request

        Message request = requestContext.getMessage();
        String orderNumber = request.evaluateString(Constants.ORDER_NUMBER);

        CustomerOrder customerOrder;
        try {
            customerOrder = customerOrderDAO.getCustomerOrder(orderNumber);
            customerOrder.setStatus(Status.SCHEDULED);
        } catch (DAOException de) {
            throw new WarehouseError(de.getMessage());
        }

        // now we schedule order for shipment with the shipping processor.
        // to demonstrate a potentially long running exchange, e.g. order
        // scheduling could require human intervention via a web module
        // interface, etc ... we do not return response here, but instead
        // save msg exchange instance for later completion, i.e. within
        // the confirm order method

        try {
            OutstandingRequest or = new OutstandingRequest(myOperation);
            customerOrder.addOutstandingRequest(or);
            customerOrderDAO.updateCustomerOrder(customerOrder);
            shippingProcessor.scheduleOrder(customerOrder);
        } catch (DAOException de) {
            throw new WarehouseError(de.getMessage());
        }

        // note that in this case, partner application should expect a
        // non-anonymous response, i.e. via ws-addressing. this is,
        // however, a binding configuration which should not influence
        // application programming logic/style

    }

    /**
     * This method maps to 'getOrderStatus' operation defined on interface implemented by this
     * web service. Invoked by <code>SpringEngine</code>. Places order by updating order
     * status to ordered and scheduling the order for shipment.
     * 
     * @param myOperation
     * @param requestContext
     */
    public void getOrderStatus(MyOperation myOperation, ActionContext requestContext) {

        // using a message property (which maps schema type to fragment
        // retreived from message using a predefined query string) we
        // extract the order number from request

        CustomerOrder customerOrder;
        try {
            Message request = requestContext.getMessage();
            String orderNumber = request.evaluateString(Constants.ORDER_NUMBER);
            customerOrder = customerOrderDAO.getCustomerOrder(orderNumber);
        } catch (DAOException de) {
            throw new WarehouseError(de.getMessage());
        }

        // create a runtime response message instance. the definition for
        // which is provided within the example-business-docs deployment
        // descriptor, i.e. we don't have to build it up manually

        ActionContext response = myOperation.createResponseAction();
        Element payload = response.getMessage().getContent();
        payload.setAttribute("orderNumber", customerOrder.getOrderNumber());
        payload.setAttribute("status", customerOrder.getStatus().toString().toLowerCase());
        myOperation.sendResponse(response);

    }

    /**
     * Package protected method invoked by ShippingProcessor once order has been scheduled for
     * shipment.
     * 
     * @param customerOrder
     */
    void confirmOrder(CustomerOrder customerOrder) {

        // shipping processor has scheduled order and estimated
        // the shipping date. retreive the outstanding request
        // and return order response to caller

        String messageExchangeID = MyOperation.toID(myRole, Constants.SCHEDULE_ORDER);
        OutstandingRequest or = customerOrder.getOutstandingRequest(messageExchangeID);
        MyOperation myOperation = or.getMessageExchange();

        // create a runtime response message instance. the definition for
        // which is provided within the example-business-docs deployment
        // descriptor, i.e. we don't have to build it up manually

        try {
            ActionContext response = myOperation.createResponseAction();
            Element payload = response.getMessage().getContent();
            payload.setAttribute("orderNumber", customerOrder.getOrderNumber());
            String date = DateFormat.getDateInstance().format(customerOrder.getEstShipDate());
            payload.setAttribute("estimatedShipDate", date);
            myOperation.sendResponse(response);
            or.setResponseSent(true);
            customerOrderDAO.updateCustomerOrder(customerOrder);
        } catch (DAOException de) {
            throw new WarehouseError(de.getMessage());
        }

    }

    /**
     * Package protected method invoked by ShippingProcessor once order has been shipped.
     * 
     * @param customerOrder
     */
    void completeOrder(CustomerOrder customerOrder) {
        
        // invoke completeOrder method on partner application, i.e. the
        // centralized order management system, which sends a shipment
        // confirmation to customer via email and closes order

        MyApplication myApplication = MyApplicationContext.getApplication();
        PartnerApplication orderManager = myApplication.getPartner(Constants.ORDER_MANAGER);
        PartnerRole pRole = orderManager.getPartnerRole(Constants.PARTNER_ROLE);
        PartnerOperation partnerOperation = pRole.createOperation(Constants.PARTNER_OPERATION);

        // create a runtime request message instance. the definition for
        // which is provided within the example-business-docs deployment
        // descriptor, i.e. we don't have to build it up manually

        ActionContext request = partnerOperation.createRequestAction();
        Element payload = request.getMessage().getContent();
        payload.setAttribute("orderNumber", customerOrder.getOrderNumber());
        String date = DateFormat.getDateInstance().format(customerOrder.getEstShipDate());
        payload.setAttribute("actualShipDate", date);
        partnerOperation.sendRequest(request);

        // the mep is in-only, i.e. no response is expected. now
        // we set status to shipped on our system and then the order
        // is complete
        
        try {
            long estShipDate = customerOrder.getEstShipDate().getTime();
            Date actualShipDate = new Date(estShipDate + Constants.THREE_DAYS_IN_MILLISECONDS);
            customerOrder.setActualShipDate(actualShipDate);
            customerOrder.setStatus(Status.SHIPPED);
            customerOrderDAO.updateCustomerOrder(customerOrder);
        } catch (DAOException de) {
            throw new WarehouseError(de.getMessage());
        }

    }

    /*
     * dependency injector. set by spring framework
     */
    public void setShippingProcessor(ShippingProcessor shippingProcessor) {
        this.shippingProcessor = shippingProcessor;
    }

    /*
     * dependency injector. set by spring framework
     */
    public void setCustomerOrderDAO(CustomerOrderDAO customerOrderDAO) {
        this.customerOrderDAO = customerOrderDAO;
    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void init() {

        // the application may be 'wired' together by a third party
        // at deployment time. perform some sanity checks

        assert (shippingProcessor != null);
        assert (customerOrderDAO != null);

        myRole = MyApplicationContext.getApplication().getMyRole(ROLE_NAME);

    }

    /*
     * Life cycle method invoked by spring framework
     */
    public void destroy() {
    }

}
