/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.example.warehouse.manager.service.dao;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;
import javax.xml.namespace.QName;

import org.bluestemsoftware.open.eoa.example.warehouse.manager.util.OutstandingRequest;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.CustomerOrder;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.Item;
import org.bluestemsoftware.open.eoa.example.warehouse.manager.vo.CustomerOrder.Status;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplication;
import org.bluestemsoftware.specification.eoa.application.spring.MyApplicationContext;

public class CustomerOrderDAO implements DataAccessObject {

    private String orderNumberPrefix;
    private DataSource dataSource;
    
    private static final String CREATE_ORDER_TABLE = "CREATE TABLE EOA_EXAMPLES_ORDER "
        + "(order_number VARCHAR(8192) NOT NULL, "
        + "zip_code CHAR(5), "
        + "ship_cost DOUBLE, "
        + "est_ship_date BIGINT, "
        + "actual_ship_date BIGINT, "
        + "status VARCHAR(9), "
        + "PRIMARY KEY (order_number))";
    
    private static final String CREATE_ITEM_TABLE = "CREATE TABLE EOA_EXAMPLES_ORDER_ITEM "
        + "(id VARCHAR(8192) NOT NULL, "
        + "order_number VARCHAR(8192), "
        + "quantity INTEGER, "
        + "PRIMARY KEY (id, order_number))";
    
    private static final String INDEX_ITEM_TABLE = "CREATE INDEX EOA_EXAMPLES_ORDER_ITEM_IDX ON EOA_EXAMPLES_ORDER_ITEM(order_number)";
    
    private static final String CREATE_ORQ_TABLE = "CREATE TABLE EOA_EXAMPLES_ORQ "
        + "(order_number VARCHAR(8192), "
        + "outstanding_request BLOB(10K))";
    
    private static final String INDEX_OR_TABLE = "CREATE INDEX EOA_EXAMPLES_ORQ_IDX ON EOA_EXAMPLES_ORQ(order_number)";
    
    private static final String INSERT_ORDER = "INSERT INTO EOA_EXAMPLES_ORDER "
        + "(order_number, zip_code, ship_cost, est_ship_date, actual_ship_date, status) "
        + "VALUES (?, ?, ?, ?, ?, ?)";
    
    private static final String INSERT_ORDER_ITEM = "INSERT INTO EOA_EXAMPLES_ORDER_ITEM "
        + "(id, order_number, quantity) "
        + "VALUES (?, ?, ?)";
    
    private static final String INSERT_ORQ = "INSERT INTO EOA_EXAMPLES_ORQ "
        + "(order_number, outstanding_request) "
        + "VALUES (?, ?)";
    
    private static final String UPDATE_ORDER = "UPDATE EOA_EXAMPLES_ORDER "
        + "SET zip_code = ?, ship_cost = ?, est_ship_date = ?, actual_ship_date = ?, status = ? "
        + "WHERE order_number = ?";
    
    private static final String SELECT_ORDER_ITEMS = "SELECT O.zip_code, O.ship_cost, O.est_ship_date, O.actual_ship_date, O.status, OI.id, OI.quantity "
        + "FROM EOA_EXAMPLES_ORDER O, EOA_EXAMPLES_ORDER_ITEM OI "
        + "WHERE O.order_number = OI.order_number AND O.order_number = ?";
    
    private static final String SELECT_OUTSTANDING_REQUESTS = "SELECT outstanding_request "
        + "FROM EOA_EXAMPLES_ORQ "
        + "WHERE order_number = ?";
    
    /*
     * life-cycle method invoked by the spring application framework
     */
    public void init() throws DAOException {
        
        // using name of engine which implements this application
        // we derive the order prefix which indicates warehouse
        // from which order is fulfilled.
        
        MyApplication myApplication = MyApplicationContext.getApplication();
        QName engineName = myApplication.getName();
        orderNumberPrefix = engineName.getNamespaceURI().substring(25,28).toUpperCase();
        
        // create our db schema. note that this example uses
        // the alakai 'system' datasource. we recommend that you
        // instead use an external datasource for your apps. we
        // also assume the sql dialect is derby ...
        
        ResultSet rs = null;
        Connection conn = null;
        PreparedStatement stmt = null;
        try {
            conn = dataSource.getConnection();
            conn.setAutoCommit(false);
            rs = conn.getMetaData().getTables(null, null, "EOA_EXAMPLES_ORDER", new String[] { "TABLE" });
            if(!rs.next()) {
                stmt = (PreparedStatement)conn.prepareStatement(CREATE_ORDER_TABLE);
                stmt.executeUpdate();
                stmt = (PreparedStatement)conn.prepareStatement(CREATE_ITEM_TABLE);
                stmt.executeUpdate();
                stmt = (PreparedStatement)conn.prepareStatement(INDEX_ITEM_TABLE);
                stmt.executeUpdate();
                stmt = (PreparedStatement)conn.prepareStatement(CREATE_ORQ_TABLE);
                stmt.executeUpdate();
                stmt = (PreparedStatement)conn.prepareStatement(INDEX_OR_TABLE);
                stmt.executeUpdate();
            }
        } catch (Throwable th) {
            if (conn != null) {
                try {
                    conn.rollback();
                } catch (SQLException ignore) {
                }
            }
            throw new DAOException(th.getMessage());
        } finally {
            try {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException ignore) {
            }
        }
        
    }
    
    public synchronized CustomerOrder createCustomerOrder(String zipCode, Set<Item> items, double shipCost) throws DAOException {
        
        String orderNumber = createOrderNumber();
        CustomerOrder co = new CustomerOrder(orderNumber, zipCode, items, shipCost);
        
        Connection conn = null;
        PreparedStatement stmt = null;
        try {
            conn = dataSource.getConnection();
            conn.setAutoCommit(false);
            stmt = (PreparedStatement)conn.prepareStatement(INSERT_ORDER);
            stmt.setString(1, orderNumber);
            stmt.setString(2, zipCode);
            stmt.setDouble(3, shipCost);
            stmt.setLong(4, 0);
            stmt.setLong(5, 0);
            stmt.setString(6, Status.PROMISED.toString());
            stmt.executeUpdate();
            for (Item item : items) {
                stmt = (PreparedStatement)conn.prepareStatement(INSERT_ORDER_ITEM);
                stmt.setString(1, item.getId());
                stmt.setString(2, orderNumber);
                stmt.setInt(3, item.getQuantity());
                stmt.executeUpdate();
            }  
            conn.commit();
        } catch (SQLException se) {
            if (conn != null) {
                try {
                    conn.rollback();
                } catch (SQLException ignore) {
                }
            }
            throw new DAOException(se.getMessage());
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException ignore) {
            }
        }
        
        return co;
        
    }
    
    public synchronized CustomerOrder getCustomerOrder(String orderNumber) throws DAOException {
        
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try {
            conn = dataSource.getConnection();
            stmt = (PreparedStatement)conn.prepareStatement(SELECT_ORDER_ITEMS);
            stmt.setString(1, orderNumber);
            rs = stmt.executeQuery();
            String zipCode = null;
            Double shipCost = null;
            Date estShipDate = null;
            Date actualShipDate = null;
            Status status = null;
            Set<Item> orderItems = new HashSet<Item>();
            while (rs.next()) {
                zipCode = rs.getString(1);
                shipCost = rs.getDouble(2);
                long tmp = rs.getLong(3);
                if (tmp > 0) {
                    estShipDate = new Date(tmp);
                }
                tmp = rs.getLong(4);
                if (tmp > 0) {
                    actualShipDate = new Date(tmp);
                }
                String temp = rs.getString(5);
                if (!temp.equals("")) {
                    status = Status.valueOf(temp);
                }                
                String itemID = rs.getString(6);
                int quantity = rs.getInt(6);
                orderItems.add(new Item(itemID, quantity));
            }
            Map<String, OutstandingRequest> ors = new HashMap<String, OutstandingRequest>();
            stmt = (PreparedStatement)conn.prepareStatement(SELECT_OUTSTANDING_REQUESTS);
            stmt.setString(1, orderNumber);
            rs = stmt.executeQuery();
            while (rs.next()) {
                ObjectInputStream in = new ObjectInputStream(rs.getBinaryStream(1));
                OutstandingRequest or = (OutstandingRequest)in.readObject();
                ors.put(or.getID(), or);
            }
            CustomerOrder co = new CustomerOrder(orderNumber, zipCode, orderItems, shipCost);
            co.setEstShipDate(estShipDate);
            co.setActualShipDate(actualShipDate);
            co.setStatus(status);
            co.setItems(orderItems);
            co.setOutstandingRequests(ors);
            return co;
        } catch (Exception se) {
            throw new DAOException(se.getMessage());
        } finally {
            try {
                if (rs != null) {
                    rs.close();
                }
                if (stmt != null) {
                    stmt.close();
                }
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException ignore) {
            }
        }
        
    }
    
    public synchronized void updateCustomerOrder(CustomerOrder customerOrder) throws DAOException {
        
        Connection conn = null;
        PreparedStatement stmt = null;
        try {
            conn = dataSource.getConnection();
            conn.setAutoCommit(false);
            stmt = (PreparedStatement)conn.prepareStatement(UPDATE_ORDER);
            stmt.setString(1, customerOrder.getZipCode());
            stmt.setDouble(2, customerOrder.getShipCost());
            if (customerOrder.getEstShipDate() != null) {
                stmt.setLong(3, customerOrder.getEstShipDate().getTime());
            } else {
                stmt.setLong(3, 0);
            }
            if (customerOrder.getActualShipDate() != null) {
                stmt.setLong(4, customerOrder.getActualShipDate().getTime());
            } else {
                stmt.setLong(4, 0);
            }
            stmt.setString(5, customerOrder.getStatus().toString());
            stmt.setString(6, customerOrder.getOrderNumber());
            stmt.executeUpdate();
            for (OutstandingRequest or : customerOrder.getOutstandingRequests()) {
                stmt = (PreparedStatement)conn.prepareStatement(INSERT_ORQ);
                stmt.setString(1, customerOrder.getOrderNumber());
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                ObjectOutputStream out = new ObjectOutputStream(baos);
                out.writeObject(or);
                byte[] bytes = baos.toByteArray();
                ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
                stmt.setBinaryStream(2, bais, bytes.length);
                stmt.executeUpdate();
            }  
            conn.commit();
        } catch (Exception se) {
            if (conn != null) {
                try {
                    conn.rollback();
                } catch (SQLException ignore) {
                }
            }
            throw new DAOException(se.getMessage());
        } finally {
            try {
                if (stmt != null) {
                    stmt.close();
                }
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException ignore) {
            }
        }
        
    }
    
    /*
     * dependency injector. set by spring framework
     */
    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }
    
    private String createOrderNumber() {
        
        // the example uses a 'standard' order number format to
        // demonstrate the use of message properties. it consists
        // of three chars followed by a dash and exactly five
        // digits ... 
        
        double random = (Math.random() * Integer.parseInt("99998"));
        String digits = Long.toString(Math.round(random));
        while(digits.length() < 5) {
            digits = "0" + digits;
        }
        
        return orderNumberPrefix + "-" + digits;
        
    }
    
}
